using UnityEngine;
using System;
using System.Collections.Generic;

namespace PerPixelDestruction
{
    public class Mod
    {
        public static void Main()
        {
            AddPPDItem("Brick Cube", "PPD Brick Cube", "brick cube.png", "brick cube view.png");
            AddPPDItem("Brick Wall", "PPD Brick Wall", "brick wall.png", "brick wall view.png");
            AddPPDItem("Wooden Pillar", "PPD Wooden Pillar", "wooden pillar.png", "wooden pillar view.png");
            AddPPDItem("Concrete slab", "PPD Concrete Slab", "concrete slab.png", "concrete slab view.png");
            AddPPDItem("Wooden Chair", "PPD Wooden Chair", "chair.png", "chair view.png");
            AddPPDItem("Wooden Table", "PPD Wooden Table", "table.png", "table view.png");
        }

        private static void AddPPDItem(string originalItem, string name, string spritePath, string viewSpritePath)
        {
            ModAPI.Register(
             new Modification()
             {
                 OriginalItem = ModAPI.FindSpawnable(originalItem),
                 NameOverride = name,
                 DescriptionOverride = "Destructible version of \"" + originalItem + "\"",
                 CategoryOverride = ModAPI.FindCategory("Misc."),
                 ThumbnailOverride = ModAPI.LoadSprite(viewSpritePath, 1, false),
                 AfterSpawn = (Instance) =>
                 {
                     PPDCreator.ConvertToDestructible(Instance, ModAPI.LoadSprite(spritePath));
                 }
             }
         );
        }
    }

    internal struct PPDCreator
    {
        internal static void ConvertToDestructible(GameObject gm, Sprite sprite)
        {
            var spriteRenderer = gm.GetComponent<SpriteRenderer>();
            spriteRenderer.sprite = sprite;
            //TODO eigenlijk moet alles kapot worden gemaakt behalve colliders, rigidbodies, en physicalbehaviours
            UnityEngine.Object.Destroy(gm.GetComponent<DestroyableBehaviour>());

            PPDestructibeBehaviour ppd = gm.AddComponent<PPDestructibeBehaviour>();

            var phys = gm.GetComponent<PhysicalBehaviour>();

            ppd.Softness = phys.Properties.Softness;
            ppd.OriginalObjectMass = phys.rigidbody.mass;
            ppd.OriginalObjectSize = new Vector2Int(sprite.texture.width, sprite.texture.height);

            ppd.Build();
        }
    }

    internal struct PPDColliderGenerator
    {
        public static IEnumerable<RectInt> GenerateColliders(Cell[][] cells, int width, int height)
        {
            while (true)
            {
                var pos = GetNextUnmarkedCell(cells, width, height);
                if (pos.x == -1)
                    break;

                var rect = CalculateNextRectAt(pos, cells, width, height);
                yield return rect;
            }
        }

        private static RectInt CalculateNextRectAt(Vector2Int pos, Cell[][] cells, int width, int height)
        {
            RectInt current = new RectInt(pos.x, pos.y, 1, 1);

            int endX = pos.x;
            for (int x = pos.x; x < width; x++)
            {
                var c = cells[x][pos.y];
                if (c.Destroyed || c.MarkedForCollision)
                    break;
                endX = x + 1;
            }

            if (endX == width - 1)
                endX = width - 1;
            current.xMax = endX;

            for (int x = pos.x; x < endX; x++)
                cells[x][pos.y].MarkedForCollision = true;

            for (int y = pos.y + 1; y < height; y++)
            {
                if (IsValidRow(cells, y, pos.x, endX))
                {
                    current.yMax = y + 1;
                    for (int x = pos.x; x < endX; x++)
                        cells[x][y].MarkedForCollision = true;
                }
                else
                    break;
            }

            return current;
        }

        private static bool IsValidRow(Cell[][] cells, int y, int startX, int endX)
        {
            for (int x = startX; x < endX; x++)
            {
                var c = cells[x][y];
                if (c.MarkedForCollision || c.Destroyed)
                    return false;
            }
            return true;
        }

        private static Vector2Int GetNextUnmarkedCell(Cell[][] cells, int width, int height)
        {
            for (int y = 0; y < height; y++)
                for (int x = 0; x < width; x++)
                {
                    var c = cells[x][y];
                    if (!c.MarkedForCollision && !c.Destroyed)
                        return new Vector2Int(x, y);
                }

            return new Vector2Int(-1, -1);
        }
    }

    public class PPDestructibeBehaviour : MonoBehaviour
    {
        private SpriteRenderer spriteRenderer;
        private PhysicalBehaviour physicalBehaviour;
        private bool initialised = false;

        private Cell[][] cells;
        private Texture2D texture;
        private int cWidth;
        private int cHeight;
        private short currentIslandCount;

        private bool dirtyCells = false;

        public float Softness = 1;
        public float OriginalObjectMass;
        public Vector2Int OriginalObjectSize;

        private float MassFraction => OriginalObjectMass * (((float)cWidth * cHeight) / ((float)OriginalObjectSize.x * OriginalObjectSize.y));

        private const float pixelSize = 1 / 35f;

        private void Awake()
        {
            spriteRenderer = GetComponent<SpriteRenderer>();
            physicalBehaviour = GetComponent<PhysicalBehaviour>();

            physicalBehaviour.OverrideShotSounds = Array.Empty<AudioClip>();
        }

        public void Build()
        {
            if (initialised) return;

            var sprite = spriteRenderer.sprite;
            texture = sprite.texture;

            cWidth = texture.width;
            cHeight = texture.height;

            Debug.Log("Building PPD object with W" + cWidth + " H" + cHeight);

            cells = new Cell[cWidth][];
            for (int i = 0; i < cWidth; i++)
                cells[i] = new Cell[cHeight];

            var pixels = texture.GetPixels32(0);

            int index = 0;

            for (int y = 0; y < cHeight; y++)
            {
                for (int x = 0; x < cWidth; x++)
                {
                    var p = pixels[index];
                    cells[x][y] = new Cell
                    {
                        Destroyed = p.a == 0,
                        Color = p,
                        IslandID = -1
                    };
                    index++;
                }
            }

            var m = MassFraction;// cWidth * cHeight / 150f;
            physicalBehaviour.rigidbody.mass = m;
            physicalBehaviour.TrueInitialMass = m;
            physicalBehaviour.InitialMass = m;

            CreateCollider();
            CalculateIslands();

            initialised = true;
        }

        private void FixedUpdate()
        {
            if (dirtyCells)
            {
                dirtyCells = false;
                CalculateIslands();
                SeparateObjectByIsland();
            }

            //if (physicalBehaviour.OnFire && UnityEngine.Random.value > 0.995)
            //{
            //    CauseDamageAtCellPoint(
            //        UnityEngine.Random.Range(0, cWidth),
            //        UnityEngine.Random.Range(0, cHeight),
            //        UnityEngine.Random.Range(1, 3)
            //        );
            //}
        }

        private void OnCollisionEnter2D(Collision2D coll)
        {
            var c = coll.GetContact(0);
            var i = c.normalImpulse;
            if (float.IsNaN(c.normalImpulse) || i < 65 - Softness * 45) return;

            Debug.Log("Collision force = " + i);

            int d = Mathf.RoundToInt(i / 40f);
            d = Mathf.Clamp(d, 1, Mathf.Min(cWidth, cHeight) / 2);

            var localPos = transform.InverseTransformPoint(c.point);
            CauseDamageAtLocalPoint(localPos, d);
        }

        private void Shot(Shot shot)
        {
            var worldPos = shot.point;
            var localPos = transform.InverseTransformPoint(worldPos);

            int d = Mathf.RoundToInt(shot.damage / (50f - Softness * 20));

            CauseDamageAtLocalPoint(localPos, Mathf.Clamp(d, 2, 7));
        }

        private void ExitShot(Shot shot)
        {
            Shot(shot);
        }

        public void CauseDamageAtLocalPoint(Vector3 point, int radiusInPixels = 4)
        {
            var cellPos = LocalPositionToCellPosition(point.x, point.y);

            CauseDamageAtCellPoint(cellPos.x, cellPos.y, radiusInPixels);
        }

        public void CauseDamageAtCellPoint(int xPos, int yPos, int radiusInPixels = 4)
        {
            for (int y = 0; y < cHeight; y++)
                for (int x = 0; x < cWidth; x++)
                {
                    var d = Mathf.RoundToInt(Vector2.Distance(new Vector2(xPos, yPos), new Vector2(x, y)));
                    if (d == radiusInPixels)
                        cells[x][y].Destroyed = true;
                }

            dirtyCells = true;
        }

        private Vector2Int LocalPositionToCellPosition(float x, float y)
        {
            x -= pixelSize / 2;
            y -= pixelSize / 2;

            var actualWidth = pixelSize * cWidth;
            var actualHeight = pixelSize * cHeight;

            x /= actualWidth;
            y /= actualHeight;

            x += 0.5f;
            y += 0.5f;

            int iX = Mathf.RoundToInt(x * cWidth);
            int iY = Mathf.RoundToInt(y * cHeight);

            return new Vector2Int(iX, iY);
        }

        private Vector2 CellPositionToLocalPosition(float x, float y)
        {
            var actualWidth = pixelSize * cWidth;
            var actualHeight = pixelSize * cHeight;
            var xRatio = x / cWidth - 0.5f;
            var yRatio = y / cHeight - 0.5f;

            return new Vector2(
                xRatio * actualWidth + pixelSize / 2,
                yRatio * actualHeight + pixelSize / 2);
        }

        private void SeparateObjectByIsland()
        {
            if (currentIslandCount == 1)
                return;

            foreach (Transform item in transform)
                Destroy(item.gameObject);

            for (short i = 0; i < currentIslandCount; i++)
                CreateNewPart(i);

            foreach (var item in GetComponents<Component>())
                if (!(item is Transform) && !(item is DeregisterBehaviour))
                    Destroy(item);
        }

        private void CreateNewPart(short islandID)
        {
            int minx = int.MaxValue;
            int miny = int.MaxValue;
            int maxx = int.MinValue;
            int maxy = int.MinValue;

            int w = 0;
            int h = 0;
            int cellCount = 0;
            int startX = -1;
            int startY = -1;

            for (int y = 0; y < cHeight; y++)
                for (int x = 0; x < cWidth; x++)
                {
                    var c = cells[x][y];
                    if (c.Destroyed || c.IslandID != islandID)
                        continue;

                    if (startX == -1) startX = x;
                    if (startY == -1) startY = y;

                    minx = Mathf.Min(minx, x);
                    miny = Mathf.Min(miny, y);

                    maxx = Mathf.Max(maxx, x);
                    maxy = Mathf.Max(maxy, y);

                    cellCount++;
                }

            maxx++;
            maxy++;

            w = Mathf.Abs(maxx - minx);
            h = Mathf.Abs(maxy - miny);

            if (w * h <= 3)
                return;

            Vector2 center = new Vector2(minx + (w - 1) / 2f, miny + (h - 1) / 2f);
            center = CellPositionToLocalPosition(center.x, center.y);

            Texture2D newTexture = new Texture2D(w, h, TextureFormat.RGBA32, false);
            newTexture.filterMode = FilterMode.Point;

            var newTextureData = newTexture.GetRawTextureData<Color32>();
            int pixelIndex = 0;
            for (int y = miny; y < maxy; y++)
                for (int x = minx; x < maxx; x++)
                {
                    var c = cells[x][y];

                    if (c.Destroyed || c.IslandID != islandID)
                        newTextureData[pixelIndex] = (Color32)Color.clear;
                    else
                        newTextureData[pixelIndex] = (Color32)c.Color;

                    pixelIndex++;
                }

            newTexture.Apply();
            bool isDebris = w * h <= 37;

            Sprite sprite = Sprite.Create(newTexture, new Rect(0, 0, newTexture.width, newTexture.height), new Vector2(.5f, .5f), 35);

            var newPart = ModAPI.CreatePhysicalObject("PPD obj island " + islandID, sprite);

            if (isDebris) //dit moet eerder gebeuren omdat unity eng is
            {
                newPart.AddComponent<DebrisComponent>();
                newPart.layer = LayerMask.NameToLayer("Debris");
            }

            if (GetComponent<FreezeBehaviour>() != null)
                newPart.AddComponent<FreezeBehaviour>();

            //unity is eng, ok? ik weet niet wat daar allemaal achter de schermen gebeurt dus je kan maar beter voorzichtig zijn en isDebris tienduizend keer testen
            if (isDebris)
            {
                newPart.transform.position = transform.TransformPoint(center);
                newPart.transform.rotation = transform.rotation;
                newPart.transform.localScale = transform.lossyScale;
            }
            else
            {
                newPart.transform.SetParent(transform);
                newPart.transform.position = transform.TransformPoint(center);
                newPart.transform.rotation = transform.rotation;
                newPart.transform.localScale = Vector2.one;
            }

            PPDestructibeBehaviour ppd = newPart.AddComponent<PPDestructibeBehaviour>();

            ppd.OriginalObjectMass = physicalBehaviour.rigidbody.mass;
            ppd.OriginalObjectSize = new Vector2Int(cWidth, cHeight);
            ppd.Softness = Softness;

            var phys = ppd.GetComponent<PhysicalBehaviour>();
            phys.Properties = physicalBehaviour.Properties;
            phys.burnIntensity = physicalBehaviour.burnIntensity;
            phys.BurnProgress = physicalBehaviour.BurnProgress;

            var myVelAtPoint = physicalBehaviour.rigidbody.GetPointVelocity(newPart.transform.position);
            var rb = phys.rigidbody;
            rb.AddForce(myVelAtPoint, ForceMode2D.Impulse);
            rb.AddTorque(phys.rigidbody.angularVelocity, ForceMode2D.Impulse);

            if (isDebris) //alweer? wat raar...
            {
                rb.AddForce(UnityEngine.Random.insideUnitCircle * 10, ForceMode2D.Impulse);
                rb.AddTorque(UnityEngine.Random.Range(-10, 10));
            }

            newPart.AddComponent<AudioSourceTimeScaleBehaviour>();

            ppd.Build();
        }

        private void CreateCollider()
        {
            var colliders = GetComponentsInChildren<Collider2D>();
            foreach (var item in colliders)
                Destroy(item);

            var actualWidth = pixelSize * cWidth;
            var actualHeight = pixelSize * cHeight;
            var actualHalfSize = new Vector2(actualWidth, actualHeight) / 2;

            foreach (var rect in PPDColliderGenerator.GenerateColliders(cells, cWidth, cHeight))
            {
                var box = gameObject.AddComponent<BoxCollider2D>();
                box.size = new Vector2(rect.size.x, rect.size.y) * pixelSize;
                box.offset = rect.center * pixelSize - actualHalfSize;
            }
        }

        private void CalculateIslands()
        {
            Debug.Log("Island calculation started...");

            ClearIslands();

            short cId = 0;
            while (true)
            {
                var pos = GetNextUnmarkedPos();

                if (pos.x == -1)
                    break;
                RecursiveFloodfill(cId, pos.x, pos.y);

                cId++;
            }

            currentIslandCount = cId;

            Debug.Log(currentIslandCount + " island(s) found");
        }

        private void ClearIslands()
        {
            for (int y = 0; y < cHeight; y++)
                for (int x = 0; x < cWidth; x++)
                    cells[x][y].IslandID = -1;
        }

        private Vector2Int GetNextUnmarkedPos()
        {
            for (int y = 0; y < cHeight; y++)
                for (int x = 0; x < cWidth; x++)
                {
                    var c = cells[x][y];
                    if (!c.Destroyed && c.IslandID == -1)
                        return new Vector2Int(x, y);
                }

            return new Vector2Int(-1, -1);
        }

        private void RecursiveFloodfill(short id, int x, int y)
        {
            if (x < 0 || x >= cWidth || y < 0 || y >= cHeight)
                return;

            var c = cells[x][y];
            if (c.Destroyed || c.IslandID == id)
                return;

            cells[x][y].IslandID = id;

            if (x > 0)
                RecursiveFloodfill(id, x - 1, y);

            if (x < cWidth - 1)
                RecursiveFloodfill(id, x + 1, y);

            if (y > 0)
                RecursiveFloodfill(id, x, y - 1);

            if (y < cHeight - 1)
                RecursiveFloodfill(id, x, y + 1);
        }

        //private void ReplaceUndoEntryWith(GameObject createdParts)
        //{
        //    if (UndoControllerBehaviour.FindRelevantAction(transform.root.gameObject, out var undoable))
        //    {
        //        switch (undoable)
        //        {
        //            case ObjectCreationAction createAction:
        //                createAction.RelevantObject = createdDebris;
        //                break;
        //        }
        //    }
        //}
    }

    public struct Cell
    {
        public bool Destroyed;
        public Color Color;

        public short IslandID;
        public bool MarkedForCollision;
    }
}